# src/__init__.py

# Importing functions from various modules
from src.nmal import decimal_to_nlgmal, nlgmal_to_decimal
from src.nlg256 import generate_nlgdecimal_constants, nlg256
from src.neccak import generate_rotation_offsets, neccak256
from src.nlgsecp256k1 import (
    mod_inverse, point_add, scalar_mult,
    generate_keypair, private_key_to_nlg256,
    public_key_to_nlg256, address_from_public_key
)

__all__ = [
    "decimal_to_nlgmal",
    "nlgmal_to_decimal",
    "generate_nlgdecimal_constants",
    "nlg256",
    "generate_rotation_offsets",
    "neccak256",
    "mod_inverse",
    "point_add",
    "scalar_mult",
    "generate_keypair",
    "private_key_to_nlg256",
    "public_key_to_nlg256",
    "address_from_public_key",
]



# secp256k1_nlg.py
# license: MIT
# This script implements secp256k1 elliptic curve key pair generation using the custom NLGdecimal number system.

import random
from src.nmal import decimal_to_nlgmal
from src.neccak import neccak256

# secp256k1 curve parameters
CURVE_P = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F
CURVE_A = 0
CURVE_B = 7
CURVE_Gx = 0x79BE667EF9DCBBAC55A06295CE870B70A0C6F14F1E8F2F0D5E4E8D0F4D8C8A8
CURVE_Gy = 0x483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B
CURVE_N = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141
CURVE_H = 1

def mod_inverse(a, p):
    """Compute the modular inverse of a modulo p."""
    return pow(a, p - 2, p)

def point_add(P, Q, curve_p):
    """Add two points P and Q on the secp256k1 curve."""
    if P == (0, 0):
        return Q
    if Q == (0, 0):
        return P
    (x1, y1) = P
    (x2, y2) = Q
    if x1 == x2 and y1 != y2:
        return (0, 0)
    if x1 == x2:
        m = (3 * x1 * x1 + CURVE_A) * mod_inverse(2 * y1, curve_p) % curve_p
    else:
        m = (y2 - y1) * mod_inverse(x2 - x1, curve_p) % curve_p
    x3 = (m * m - x1 - x2) % curve_p
    y3 = (m * (x1 - x3) - y1) % curve_p
    return (x3, y3)

def scalar_mult(k, point, curve_p):
    """Multiply a point by a scalar k on the secp256k1 curve."""
    result = (0, 0)
    addend = point
    while k:
        if k & 1:
            result = point_add(result, addend, curve_p)
        addend = point_add(addend, addend, curve_p)
        k >>= 1
    return result

def generate_keypair():
    """Generate a secp256k1 key pair."""
    priv_key = random.getrandbits(256) % CURVE_N
    pub_key = scalar_mult(priv_key, (CURVE_Gx, CURVE_Gy), CURVE_P)
    return priv_key, pub_key

def private_key_to_nlg256(priv_key):
    """Convert a private key to NLG256 format."""
    return decimal_to_nlgmal(priv_key)

def public_key_to_nlg256(pub_key):
    """Convert a public key to NLG256 format."""
    x, y = pub_key
    return decimal_to_nlgmal(x) + decimal_to_nlgmal(y)

def address_from_public_key(pub_key):
    """Generate an address from a public key."""
    pub_key_bytes = pub_key[0].to_bytes(32, 'big') + pub_key[1].to_bytes(32, 'big')
    hashed_pub_key = neccak256(pub_key_bytes)
    address = hashed_pub_key[-40:]  # Take the last 20 bytes (40 hex characters)
    return address





import random
import struct

from nmal import *


def generate_nlgdecimal_constants(size, bit_length=64):
    """Generate a list of random NLGdecimal constants."""
    constants = []
    for _ in range(size):
        value = random.getrandbits(bit_length)  # Random integer with specified bit length
        nlgdecimal_str = decimal_to_nlgmal(value)
        constants.append(nlgdecimal_str)
    return constants

def generate_rotation_offsets(size, max_offset):
    """Generate rotation offsets within a valid range."""
    return [random.randint(0, max_offset) for _ in range(size)]

# Generate Neccak-256 constants dynamically
RHO_OFFSETS_NLG = generate_rotation_offsets(24, 63)  # Rho offsets within 0-63 for 64-bit rotations
PI_NLG = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24]  # Linear Pi permutation values
THETA_NLG = generate_nlgdecimal_constants(24)  # Theta transformation, 24 constants

def rotr(n, x, bits=64):
    """Rotate right."""
    return ((x >> n) | (x << (bits - n))) & ((1 << bits) - 1)

def neccak_f(state):
    # Neccak-f permutation
    for round_idx in range(24):
        # Theta step
        C = [0] * 5
        D = [0] * 5
        for x in range(5):
            C[x] = state[x] ^ state[x + 5] ^ state[x + 10] ^ state[x + 15] ^ state[x + 20]
        for x in range(5):
            D[x] = C[(x - 1) % 5] ^ rotr(1, C[(x + 1) % 5], 64)
        for x in range(25):
            state[x] ^= D[x % 5]

        # Rho and Pi steps
        B = [0] * 25
        for x in range(5):
            for y in range(5):
                rho_offset = RHO_OFFSETS_NLG[(x + 3 * y) % 24]
                B[((PI_NLG[x] * 5) + y) % 25] = rotr(rho_offset, state[x * 5 + y])

        # Chi step
        for x in range(5):
            for y in range(5):
                state[x * 5 + y] = B[x * 5 + y] ^ ((~B[(x + 1) % 5 * 5 + y]) & B[(x + 2) % 5 * 5 + y])

        # Iota step
        state[0] ^= nlgmal_to_decimal(THETA_NLG[round_idx])

def neccak256(data):
    # Initialize state
    state = [0] * 25
    data = bytearray(data)
    data_len = len(data) * 8

    # Padding
    data.append(0x01)
    while len(data) % 200 != 192:
        data.append(0x00)
    data += struct.pack('<Q', data_len)

    # Absorb phase
    for i in range(0, len(data), 200):
        block = data[i:i + 200]
        for j in range(25):
            state[j] ^= struct.unpack('<Q', block[j * 8:(j + 1) * 8])[0]
        neccak_f(state)

    # Squeeze phase
    output = bytearray()
    for _ in range(2):  # Two 64-byte blocks for Neccak-256
        for j in range(25):
            output += struct.pack('<Q', state[j])
        neccak_f(state)

    return ''.join(decimal_to_nlgmal(x) for x in struct.unpack('<32B', output[:32]))  # Return first 256 bits as NLGdecimal


# nlg256.py
# license: MIT
# This script implements a NLG-256 hashing algorithm using the custom NLGdecimal number system.

import struct
import random
from nmal import *


def generate_nlgdecimal_constants(size):
    """Generate a list of random NLGdecimal constants."""
    constants = []
    for _ in range(size):
        value = random.getrandbits(32)  # 32-bit random integer
        nlgdecimal_str = decimal_to_nlgmal(value)
        constants.append(nlgdecimal_str)
    return constants


# Generate nlg-256 constants dynamically
K_NLG = generate_nlgdecimal_constants(64)


# Helper functions
def rotr(n, x):
    return (x >> n) | (x << (32 - n)) & 0xFFFFFFFF


def nlg_compress(state, block):
    W = [0] * 64
    for i in range(16):
        W[i] = struct.unpack('>I', block[i * 4:(i + 1) * 4])[0]

    for i in range(16, 64):
        s0 = rotr(7, W[i - 15]) ^ rotr(18, W[i - 15]) ^ (W[i - 15] >> 3)
        s1 = rotr(17, W[i - 2]) ^ rotr(19, W[i - 2]) ^ (W[i - 2] >> 10)
        W[i] = (W[i - 16] + s0 + W[i - 7] + s1) & 0xFFFFFFFF

    a, b, c, d, e, f, g, h = state

    for i in range(64):
        s1 = rotr(6, e) ^ rotr(11, e) ^ rotr(25, e)
        ch = (e & f) ^ (~e & g)
        temp1 = (h + s1 + ch + nlgmal_to_decimal(K_NLG[i]) + W[i]) & 0xFFFFFFFF
        s0 = rotr(2, a) ^ rotr(13, a) ^ rotr(22, a)
        maj = (a & b) ^ (a & c) ^ (b & c)
        temp2 = (s0 + maj) & 0xFFFFFFFF

        h = g
        g = f
        f = e
        e = (d + temp1) & 0xFFFFFFFF
        d = c
        c = b
        b = a
        a = (temp1 + temp2) & 0xFFFFFFFF

    state[0] = (state[0] + a) & 0xFFFFFFFF
    state[1] = (state[1] + b) & 0xFFFFFFFF
    state[2] = (state[2] + c) & 0xFFFFFFFF
    state[3] = (state[3] + d) & 0xFFFFFFFF
    state[4] = (state[4] + e) & 0xFFFFFFFF
    state[5] = (state[5] + f) & 0xFFFFFFFF
    state[6] = (state[6] + g) & 0xFFFFFFFF
    state[7] = (state[7] + h) & 0xFFFFFFFF


def nlg256(data):
    data = bytearray(data)
    length = struct.pack('>Q', 8 * len(data))
    data.append(0x80)
    while (len(data) + 8) % 64 != 0:
        data.append(0)
    data += length

    state = [
        0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a,
        0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19,
    ]

    for i in range(0, len(data), 64):
        nlg_compress(state, data[i:i+64])

    return ''.join(decimal_to_nlgmal(x) for x in state)






# nmal.py
# license: MIT
# This script implements a custom number system named "NLG DECIMAL" (shortened to "NLGmal").
# The NLGmal number system uses the characters 0-9 and m-z, making it a base-24 system.
# This script includes functions to convert decimal numbers to NLGmal and vice versa.

# Defining the NLGmal characters
NLGmal_chars = "0123456789mnopqrstuvwxyz"

base = len(NLGmal_chars)

# Dictionary for quick look-up of character values
char_to_value = {char: idx for idx, char in enumerate(NLGmal_chars)}


def decimal_to_nlgmal(decimal):
    """Convert a decimal number to an NLGmal number."""
    if decimal == 0:
        return NLGmal_chars[0]

    nlgmal = ""
    while decimal > 0:
        nlgmal = NLGmal_chars[decimal % base] + nlgmal
        decimal //= base
    return nlgmal


def nlgmal_to_decimal(nlgmal):
    """Convert an NLGmal number to a decimal number."""
    # Check for capital letters and issue a warning
    if any(char.isupper() for char in nlgmal):
        print("Warning: Capital letters detected. Its good to use small small letter. Please be carefull next time :)")
        nlgmal = nlgmal.lower()

    decimal = 0
    for char in nlgmal:
        decimal = decimal * base + char_to_value[char]
    return decimal


